import random
from copy import copy, deepcopy

from Player import Player
from Choice import Choice, Move
from ZeroSum import solveZeroSumMatrix
from DumbUI import DumbShowUI

class Robot(Player):
    shortDesc = "???"

class RandomRobot(Robot):
    shortDesc = "plays entirely randomly"
    def getMove(self, pl):
	return random.choice(list(self.state.possMoves(pl)))
    def getChoice(self,pl):
	primary = self.getMove(pl)
	choice = Choice(primary, None)
	while not self.state.checkLegalChoice(choice, pl):
	    choice = Choice(primary,self.getMove(pl))
	return choice

class StupidRobot(DumbShowUI):
    shortDesc = "Low depth brute-forcing with heuristically chosen considered moves"
    MATERIAL_VALUES = {'Pawn':1, 'Knight':3, 'Bishop':3, 'Queen':8, 'Rook':5,
	    'King':1000}
    def __init__(self, socket, isServer, gameOpts, opts):
	Player.__init__(self, socket, isServer, gameOpts, opts)
	try: self.depth = int(opts["robotParams"])
	except KeyError, ValueError: self.depth = 3
	self.maxBreadth = 5
	self.pcs = list(self.state.possCoords())
    def getChoice(self, forpl):
	def evalNode(depth,value):
	    if depth == self.depth or abs(value) > 900:
		return value,[],[]
	    # narrow as we go up
	    maxBreadth = max(2, self.maxBreadth - depth)
	    def obvMoves():
		moves = [[],[]]
		poss = map(lambda pl: self.state.possMoves(pl), [0,1])
		lposs = [[],[]]
		for pl in [0,1]:
		    for move in poss[pl]:
			if move.frm == move.to:
			    # resignation
			    continue
			lposs[pl].append(move)
			if self.state.threat(move,pl):
			    taken = self.state.pieceIn(move.to)
			    important = taken and self.MATERIAL_VALUES[taken.type] >= 5
			    if self.state.threatIsEstablished(move, pl):
				if important:
				    moves[pl].insert(0,move)
				else:
				    moves[pl].append(move)
			    random.shuffle(self.pcs)
			    for to in self.pcs:
				flee = Move(move.to,to)
				if self.state.checkLegalSimulMove(flee,1-pl):
				    if important:
					moves[1-pl].insert(0,flee)
				    else:
					moves[1-pl].append(flee)
				    break
			    if important:
				# TODO: try a blocking move too
				# TODO: really we ought to try various
				#  defensive moves until we find one which
				#  "works" - this would require some cunning
				#  cheap lookahead.
				pass
			    if sum([ len(moves[p]) > maxBreadth
				    for p in [0,1] ]) == 2:
				break
		for pl in [0,1]:
		    moves[pl] = moves[pl][:maxBreadth]
		    if len(moves[pl]) < maxBreadth:
			for move in poss[pl]:
			    lposs[pl].append(move)
			if lposs[pl]:
			    random.shuffle(lposs[pl])
			    while maxBreadth > len(moves[pl]):
				try:
				    m = lposs[pl].pop()
				except IndexError:
				    break
				if m not in moves[pl]:
				    moves[pl].append(m)
		return moves
	    obvMvs = obvMoves()
	    obvChoices = [[],[]]
	    for pl in [0,1]:
		for move in obvMvs[pl]:
		    choice = Choice(move)
		    def getSecondary():
			mvs = obvMvs[pl]
			random.shuffle(mvs)
			for secondary in mvs:
			    ch = Choice(move, secondary)
			    if self.state.checkLegalChoice(ch, pl):
				return ch
			mvs = list(self.state.possMoves(pl))
			random.shuffle(mvs)
			for secondary in mvs:
			    ch = Choice(move, secondary)
			    if self.state.checkLegalChoice(ch, pl):
				return ch
			return None
		    if not self.state.checkLegalChoice(choice, pl):
			choice = getSecondary()
		    if choice:
			obvChoices[pl].append(choice)
		if obvChoices[pl] == []:
		    resignation = self.state.resignMove(pl)
		    if resignation:
			obvChoices[pl].append(Choice(resignation))
	    if obvChoices[0] == obvChoices[1] == []:
		# no kings left! It's a draw!
		return 0, [], []

	    matrix = []
	    for ours in obvChoices[forpl]:
		vals = []
		for theirs in obvChoices[1-forpl]:
		    if forpl == 0: chs = [ours,theirs]
		    else: chs = [theirs,ours]
		    print depth, chs[0], chs[1]
		    oldUpdates = self.state.update(chs)
		    dvalue = sum([
			(-1)**(rem.__class__ == self.state.PlacePiece) *
			(-1)**(forpl==rem.piece.player) *
			self.MATERIAL_VALUES[rem.piece.type]
			for rem in self.state.lastUpdates
			if rem.__class__ in
			    (self.state.RemovePiece, self.state.PlacePiece) ])
		    v = (evalNode(depth+1, value + dvalue)[0])
		    vals.append(v)
		    print depth, chs[0], chs[1], v
		    self.state.undoTurn(oldUpdates)
		matrix.append(vals)
	    #print matrix
	    v, ps = solveZeroSumMatrix(matrix)
	    #print v, ps, map(str,obvChoices[forpl])
	    return v, ps, obvChoices[forpl]
	v, ps, chs = evalNode(0,0)
	print v, ps, map(str,chs)
	r = random.random()
	while r > ps[0]:
	    r -= ps[0]
	    del ps[0]
	    del chs[0]
	return chs[0]
	    


robots = {
	    "random" : RandomRobot,
	    "stupid" : StupidRobot,
	}
