#!/usr/bin/env python

# Solve zero-sum games, using GLPK

# TODO: implement the algorithm in python, or provide python bindings to
#   GLPK's C API

import os,sys
from string import *

import tempfile

class ZeroSum:
    def __init__(self, matrix):
	"""matrix: square matrix of payoffs for the row player, a list of rows
	"""
	self.matrix = matrix
	self.rows = len(matrix)
	self.columns = len(matrix[0])
    def __str__(self):
	ret = '\t' + join(map(str, range(self.columns)),'\t') + '\n'
	for row in range(self.rows):
	    ret += '%d\t' % row + join(map(str,self.matrix[row]), '\t') + '\n'
	return ret
    def mathstr(self):
	ret = '\t' + join(map(str, range(self.columns)),'\t') + ':=\n'
	for row in range(self.rows):
	    ret += '%d\t' % row + join(map(str,self.matrix[row]), '\t') + '\n'
	ret += ';'
	return ret

ZERO_SUM_MATHPROG_HEADER = """
set I;
set J;
param a{I,J};
var p{i in I};
var t;
maximize mint: t;
s.t. maximin {j in J}: sum{i in I} a[i,j]*p[i] >= t;
s.t. prob: sum{i in I} p[i] == 1;
s.t. prob2 {i in I}: p[i] >= 0;
solve;
display: t, p;
"""

def solveZeroSum(zeroSum):
    assert zeroSum.rows > 0 and zeroSum.columns > 0
    datastring = "data;\n"
    datastring += "set I := %s;\n" % join(map(str, range(zeroSum.rows)), ',')
    datastring += "set J := %s;\n" % join(map(str, range(zeroSum.columns)), ',')
    datastring += "param a default 0.0:\n" + zeroSum.mathstr() + "\n"
    datastring += "end;\n"

    # FIXME: silly to use a tempfile really, but would need to deal with
    # buffering problems to use glpsol in a pipe
    tmpfile,tmpname = tempfile.mkstemp(prefix='zsum')
    os.write(tmpfile, ZERO_SUM_MATHPROG_HEADER + datastring)
    os.close(tmpfile)
    result = os.popen3('glpsol -m %s -y /dev/stderr' % tmpname)[2].read()
    #print datastring, result
    os.remove(tmpname)

    lines = result.splitlines()

    def afterEquals(line):
	i = line.find('=')
	return line[i+1:]

    if len(lines) < zeroSum.rows + 1:
	print("bad glpsol output:")
	print(lines)
	print("input:")
	print(ZERO_SUM_MATHPROG_HEADER + datastring)
	raise BaseException("bad output from glpsol")
    value = float(afterEquals(lines[1]))
    p = [ float(afterEquals(lines[i+2])) for i in range(zeroSum.rows) ]

    return value, p

def solveZeroSumMatrix(matrix):
    """Utility function
    """
    return solveZeroSum(ZeroSum(matrix))

if __name__ == '__main__':
    zeroSum = ZeroSum([
	[1,-3],
	[-3,1],
	[0,0]
	])
    print solveZeroSum(zeroSum)
